@@ -1,7 +1,14 @@
+from __future__ import annotations
+
+from typing import TYPE_CHECKING, Callable
+
+if TYPE_CHECKING:
+    from hatchling.metadata.core import ProjectMetadata
+
 DEFAULT_METADATA_VERSION = '2.1'
 
 
-def get_core_metadata_constructors():
+def get_core_metadata_constructors() -> dict[str, Callable]:
     """
     https://packaging.python.org/specifications/core-metadata/
     """
@@ -13,7 +20,7 @@ def get_core_metadata_constructors():
     }
 
 
-def construct_metadata_file_1_2(metadata, extra_dependencies=()):
+def construct_metadata_file_1_2(metadata: ProjectMetadata, extra_dependencies: tuple[str] | None = None) -> str:
     """
     https://peps.python.org/pep-0345/
     """
@@ -72,7 +79,7 @@ def construct_metadata_file_1_2(metadata, extra_dependencies=()):
     return metadata_file
 
 
-def construct_metadata_file_2_1(metadata, extra_dependencies=()):
+def construct_metadata_file_2_1(metadata: ProjectMetadata, extra_dependencies: tuple[str] | None = None) -> str:
     """
     https://peps.python.org/pep-0566/
     """
@@ -110,6 +117,9 @@ def construct_metadata_file_2_1(metadata, extra_dependencies=()):
             else:
                 metadata_file += f'{indent}{line}\n'
 
+    if metadata.core.license_expression:
+        metadata_file += f'License-Expression: {metadata.core.license_expression}\n'
+
     if metadata.core.license_files:
         for license_file in metadata.core.license_files:
             metadata_file += f'License-File: {license_file}\n'
@@ -137,7 +147,8 @@ def construct_metadata_file_2_1(metadata, extra_dependencies=()):
             metadata_file += f'Provides-Extra: {option}\n'
             for dependency in dependencies:
                 if ';' in dependency:
-                    metadata_file += f'Requires-Dist: {dependency} and extra == {option!r}\n'
+                    dep_name, dep_env_marker = dependency.split(';', maxsplit=1)
+                    metadata_file += f'Requires-Dist: {dep_name}; ({dep_env_marker.strip()}) and extra == {option!r}\n'
                 elif '@ ' in dependency:
                     metadata_file += f'Requires-Dist: {dependency} ; extra == {option!r}\n'
                 else:
@@ -150,7 +161,7 @@ def construct_metadata_file_2_1(metadata, extra_dependencies=()):
     return metadata_file
 
 
-def construct_metadata_file_2_2(metadata, extra_dependencies=()):
+def construct_metadata_file_2_2(metadata: ProjectMetadata, extra_dependencies: tuple[str] | None = None) -> str:
     """
     https://peps.python.org/pep-0643/
     """
@@ -188,6 +199,9 @@ def construct_metadata_file_2_2(metadata, extra_dependencies=()):
             else:
                 metadata_file += f'{indent}{line}\n'
 
+    if metadata.core.license_expression:
+        metadata_file += f'License-Expression: {metadata.core.license_expression}\n'
+
     if metadata.core.license_files:
         for license_file in metadata.core.license_files:
             metadata_file += f'License-File: {license_file}\n'
@@ -215,7 +229,8 @@ def construct_metadata_file_2_2(metadata, extra_dependencies=()):
             metadata_file += f'Provides-Extra: {option}\n'
             for dependency in dependencies:
                 if ';' in dependency:
-                    metadata_file += f'Requires-Dist: {dependency} and extra == {option!r}\n'
+                    dep_name, dep_env_marker = dependency.split(';', maxsplit=1)
+                    metadata_file += f'Requires-Dist: {dep_name}; ({dep_env_marker.strip()}) and extra == {option!r}\n'
                 elif '@ ' in dependency:
                     metadata_file += f'Requires-Dist: {dependency} ; extra == {option!r}\n'
                 else:
@@ -228,7 +243,7 @@ def construct_metadata_file_2_2(metadata, extra_dependencies=()):
     return metadata_file
 
 
-def construct_metadata_file_2_3(metadata, extra_dependencies=()):
+def construct_metadata_file_2_3(metadata: ProjectMetadata, extra_dependencies: tuple[str] | None = None) -> str:
     """
     https://peps.python.org/pep-0639/
     """
@@ -285,7 +300,8 @@ def construct_metadata_file_2_3(metadata, extra_dependencies=()):
             metadata_file += f'Provides-Extra: {option}\n'
             for dependency in dependencies:
                 if ';' in dependency:
-                    metadata_file += f'Requires-Dist: {dependency} and extra == {option!r}\n'
+                    dep_name, dep_env_marker = dependency.split(';', maxsplit=1)
+                    metadata_file += f'Requires-Dist: {dep_name}; ({dep_env_marker.strip()}) and extra == {option!r}\n'
                 elif '@ ' in dependency:
                     metadata_file += f'Requires-Dist: {dependency} ; extra == {option!r}\n'
                 else:
